﻿// Copyright (C) 2014 Kazuhiro Fujieda <fujieda@users.osdn.me>
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using ExpressionToCodeLib;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace KancolleSniffer.Test
{
    [TestClass]
    public class SnifferTest
    {
        public static StreamReader OpenLogFile(string name)
        {
            var dir = Path.GetDirectoryName(Path.GetDirectoryName(Environment.CurrentDirectory));
// ReSharper disable once AssignNullToNotNullAttribute
            var path = Path.Combine(dir, Path.Combine("logs", name + ".log.gz"));
            return new StreamReader(new GZipStream(File.Open(path, FileMode.Open), CompressionMode.Decompress));
        }

        public static void SniffLogFile(Sniffer sniffer, string name, Action<Sniffer> action = null)
        {
            var ln = 0;
            using (var stream = OpenLogFile(name))
            {
                while (!stream.EndOfStream)
                {
                    var triple = new List<string>();
                    foreach (var s in new[] {"url: ", "request: ", "response: "})
                    {
                        var line = stream.ReadLine();
                        ln++;
                        if (line == null)
                            throw new Exception($"ログのurl, request, responseがそろっていません: {ln:d}行目");
                        if (!line.StartsWith(s))
                            throw new Exception($"ログに不正な行が含まれています: {ln:d}行目");
                        triple.Add(line.Substring(s.Length));
                    }
                    var json = JsonParser.Parse(triple[2]);
                    sniffer.Sniff(triple[0], triple[1], json);
                    action?.Invoke(sniffer);
                }
            }
        }

        /// <summary>
        ///  演習で受けたダメージが次の戦闘の結果に影響しない
        /// </summary>
        [TestMethod]
        public void DamageInPracticeNotSpillIntoSortie()
        {
            var expected = new[] {31, 15, 15};
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "battle_001");
            var result = sniffer.GetShipStatuses(0).Select(s => s.NowHp);
            PAssert.That(() => (expected.SequenceEqual(result)));
        }

        /// <summary>
        /// 演習では大破警告を出さない
        /// </summary>
        [TestMethod]
        public void IgnoreDamagedShipsInPractice()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "practice_001");
            PAssert.That(() => !sniffer.BadlyDamagedShips.Any());
        }

        /// <summary>
        /// 夜戦のダメージを戦闘結果に反映する
        /// </summary>
        [TestMethod]
        public void CaptureDamageInNightCombat()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "battle_002");
            AssertEqualBattleResult(sniffer, new[] {28, 1, 13});
        }

        private void AssertEqualBattleResult(Sniffer sniffer, IEnumerable<int> expected, string msg = null)
        {
            var result = sniffer.GetShipStatuses(0).Select(s => s.NowHp);
            PAssert.That(() => (expected.SequenceEqual(result)), msg);
        }

        /// <summary>
        /// 連合艦隊(水上打撃部隊)による戦闘のダメージを結果に反映する
        /// </summary>
        [TestMethod]
        public void CombinedFleetSurface()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "combined_surface_001");
            AssertEauqlCombinedResult(sniffer, new[] {40, 77, 77, 33, 51, 47}, new[] {39, 35, 11, 39, 37, 40});

            SniffLogFile(sniffer, "combined_surface_002");
            AssertEauqlCombinedResult(sniffer, new[] {40, 77, 77, 33, 15, 6}, new[] {39, 35, 4, 3, 14, 40});
        }

        private void AssertEauqlCombinedResult(Sniffer sniffer, IEnumerable<int> expected0, IEnumerable<int> expected1,
            string msg = null)
        {
            var result0 = sniffer.GetShipStatuses(0).Select(s => s.NowHp);
            var result1 = sniffer.GetShipStatuses(1).Select(s => s.NowHp);
            PAssert.That(() => (expected0.SequenceEqual(result0) && expected1.SequenceEqual(result1)), msg);
        }

        /// <summary>
        /// 開幕夜戦のダメージを戦闘結果に反映する
        /// </summary>
        [TestMethod]
        public void SpMidnight()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "sp_midnight_001");
            AssertEqualBattleResult(sniffer, new[] {1});
            PAssert.That(() => sniffer.BadlyDamagedShips.Any());
        }

        /// <summary>
        /// 連合艦隊(空母機動部隊)による戦闘のダメージを結果に反映する
        /// </summary>
        [TestMethod]
        public void CombinedFleetAir()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "combined_air_001");
            AssertEauqlCombinedResult(sniffer, new[] {40, 98, 90, 66, 78, 86}, new[] {47, 41, 5, 42, 43, 29});
            PAssert.That(() => sniffer.BadlyDamagedShips.Any());

            SniffLogFile(sniffer, "combined_air_002");
            AssertEauqlCombinedResult(sniffer, new[] {13, 87, 90, 59, 69, 86}, new[] {47, 41, 5, 20, 43, 29});
        }

        /// <summary>
        /// 護衛退避を正しく処理する
        /// </summary>
        [TestMethod]
        public void EscapeShip()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "combined_escape_001");
            AssertEauqlCombinedResult(sniffer, new[] {37, 105, 106, 90, 66, 10}, new[] {41, 41, 37, 44, 43, 43},
                "連合艦隊で2戦して大破が出るまで");
            PAssert.That(() => sniffer.BadlyDamagedShips.Any());
            SniffLogFile(sniffer, "combined_escape_002");
            PAssert.That(() => sniffer.GetShipStatuses(0)[5].Escaped && sniffer.GetShipStatuses(1)[1].Escaped,
                "続けて護衛退避を実行");
            PAssert.That(() => !sniffer.BadlyDamagedShips.Any());
            SniffLogFile(sniffer, "combined_escape_003");
            AssertEauqlCombinedResult(sniffer, new[] {37, 105, 106, 90, 1, 10}, new[] {41, 41, 32, 44, 43, 43},
                "もう一戦して大破が出るまで");
            PAssert.That(() => sniffer.BadlyDamagedShips.Any());
            SniffLogFile(sniffer, "combined_escape_004");
            PAssert.That(() => sniffer.GetShipStatuses(0)[5].Escaped && sniffer.GetShipStatuses(1)[1].Escaped &&
                               sniffer.GetShipStatuses(0)[4].Escaped && sniffer.GetShipStatuses(1)[2].Escaped,
                "続けて護衛退避を実行");
            PAssert.That(() => !sniffer.BadlyDamagedShips.Any());
        }

        /// <summary>
        /// 護衛退避した空母を除いた制空値を計算する
        /// </summary>
        [TestMethod]
        public void FighterPowerWithoutEscapedShip()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "combined_escape_001");
            PAssert.That(() => sniffer.GetFighterPower(0)[0] == 187);
            SniffLogFile(sniffer, "combined_escape_002");
            PAssert.That(() => sniffer.GetFighterPower(0)[0] == 65);
        }

        /// <summary>
        /// 護衛退避した艦娘を除いて勝利判定を計算する
        /// </summary>
        [TestMethod]
        public void BattleResultRankWithEscapedShip()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "combined_escape_005");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.D);
        }


        /// <summary>
        /// 夜戦の開始時に昼戦の結果を反映する
        /// </summary>
        [TestMethod]
        public void ResultOfDayBattleShowInNightBattle()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "battle_003");
            AssertEqualBattleResult(sniffer, new[] {28, 2, 13});
            PAssert.That(() => !sniffer.BadlyDamagedShips.Any(), "夜戦の開始時は大破警告を出さない");
        }

        /// <summary>
        /// 通常艦隊の航空戦のダメージを結果に反映する
        /// </summary>
        [TestMethod]
        public void AirBattle()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "airbattle_001");
            AssertEqualBattleResult(sniffer, new[] {37, 36, 31, 37, 17, 63}, "夜戦あり");

            sniffer = new Sniffer();
            SniffLogFile(sniffer, "airbattle_002");
            AssertEqualBattleResult(sniffer, new[] {66, 36, 16, 27, 35, 38}, "昼戦のみ");
        }

        /// <summary>
        /// 支援射撃による敵の損傷を勝利判定に反映させる
        /// </summary>
        [TestMethod]
        public void SupportShellingChangeResultRank()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "support_001");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.A);
        }

        /// <summary>
        /// 敵連合艦隊への航空支援攻撃を反映させる
        /// </summary>
        [TestMethod]
        public void SupportAirAttackResultRank()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "support_002");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.P);
        }

        /// <summary>
        /// ダメコンの使用を戦闘結果に反映させる
        /// </summary>
        [TestMethod]
        public void DamageControl()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "damecon_001");
            AssertEqualBattleResult(sniffer, new[] {30, 1, 3}, "戦闘前");
            PAssert.That(() => sniffer.GetShipStatuses(0)[1].Slot.Select(item => item.Id).
                SequenceEqual(new[] {2, 4593, -1, -1, -1}), "ダメコンを二つ装備");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.P, "判定は完全勝利");
            SniffLogFile(sniffer, "damecon_002");
            AssertEqualBattleResult(sniffer, new[] {30, 1, 3}, "戦闘後");
            PAssert.That(() => sniffer.GetShipStatuses(0)[1].Slot.Select(item => item.Id).
                SequenceEqual(new[] {-1, 4593, -1, -1, -1}), "ダメコンを一つ消費");
        }

        /// <summary>
        /// 補強装備スロットのダメコンの使用を反映する
        /// </summary>
        [TestMethod]
        public void DamgeControlInSlotEx()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "damecon_003");
            PAssert.That(() => sniffer.GetShipStatuses(0)[5].SlotEx.Id == 10306, "戦闘前");
            SniffLogFile(sniffer, "damecon_004");
            PAssert.That(() => sniffer.GetShipStatuses(0)[5].SlotEx.Id == -1, "戦闘後");
        }

        /// <summary>
        /// 旗艦大破で女神を使った進撃を反映する
        /// </summary>
        [TestMethod]
        public void DamageControlForFlagShip()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "damecon_005");
            var fs = sniffer.GetShipStatuses(0)[0];
            PAssert.That(() => fs.SlotEx.Id == -1);
            PAssert.That(() => fs.NowHp == fs.MaxHp);
            PAssert.That(() => sniffer.BadlyDamagedShips.Length == 0);
        }

        /// <summary>
        /// 戦闘開始時に大破艦がいても夜戦突入時に警告を出さない
        /// </summary>
        [TestMethod]
        public void NotWarnDamagedShipOnRunIntoNightBattle()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "damecon_006");
            PAssert.That(() => sniffer.BadlyDamagedShips.Length == 0);
        }

        /// <summary>
        /// 轟沈した艦娘に大破警告を出さない
        /// </summary>
        [TestMethod]
        public void NotWarnToSunkShip()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "sunk_001");
            AssertEqualBattleResult(sniffer, new[]{26, 0, 1, 3});
            PAssert.That(() => sniffer.BadlyDamagedShips.SequenceEqual(new[] {"菊月", "雪風"}));
        }

        /// <summary>
        /// 連合艦隊(水上打撃部隊)で二回目の砲撃戦がない場合を正しく処理する
        /// </summary>
        [TestMethod]
        // ReSharper disable once InconsistentNaming
        public void CombinedFleetSurfaceWithout2ndShelling()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "combined_surface_003");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.P);
        }

        /// <summary>
        /// 連合艦隊(水上打撃部隊)で開幕雷撃で戦闘が終わる場合を正しく処理する
        /// </summary>
        [TestMethod]
        // ReSharper disable once InconsistentNaming
        public void CombinedFleetSurfaceWithoutAnyShelling()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "combined_surface_004");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.P);
        }

        /// <summary>
        /// ship2に代わるship_deckを処理する
        /// </summary>
        [TestMethod]
        public void ShipDeck()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "ship_deck_001");
            PAssert.That(() => sniffer.GetShipStatuses(0)[0].Fuel == 36);
        }

        /// <summary>
        /// 夜戦かどうかを選択する画面でリロードしても結果を次の戦闘に持ち越さない
        /// </summary>
        [TestMethod]
        public void ReloadBeforeBattleResult()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "reload_001");
            PAssert.That(() => sniffer.GetShipStatuses(0)[0].NowHp == 41);
        }

        /// <summary>
        /// 航空戦のない演習を正しく処理する
        /// </summary>
        [TestMethod]
        public void PracticeWithoutAirBattle()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "practice_002");
            PAssert.That(() => sniffer.Battle.AirControlLevel == -1);
        }

        /// <summary>
        /// 出撃時に大破している艦娘がいたら警告する
        /// </summary>
        [TestMethod]
        public void DamagedShipWarningOnMapStart()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "mapstart_001");
            PAssert.That(() => sniffer.BadlyDamagedShips.SequenceEqual(new[] {"大潮"}));
        }

        /// <summary>
        /// 連合艦隊に大破艦がいる状態で第3艦隊が出撃したときに警告しない
        /// </summary>
        [TestMethod]
        public void NotWarnDamagedShipInCombinedFleetOnMapStart()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "mapstart_002");
            PAssert.That(() => !sniffer.BadlyDamagedShips.Any());
        }

        /// <summary>
        /// 先制対潜攻撃を正しく処理する
        /// </summary>
        [TestMethod]
        public void OpeningTaisen()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "opening_taisen_001");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.P);
        }

        /// <summary>
        /// 敵の空スロットの値が-1ではなく0のときにエラーにならない
        /// </summary>
        [TestMethod]
        public void ZeroValueEmptySlot()
        {
            var sniffer = new Sniffer {UseOldEnemyId = true};
            SniffLogFile(sniffer, "battle_006");
            PAssert.That(() => sniffer.Battle.EnemyFighterPower.AirCombat == 0);
        }

        /// <summary>
        /// B勝利を正しく判定する
        /// </summary>
        [TestMethod]
        public void JudgeResultRankB()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "battle_008");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.B);
        }

        /// <summary>
        /// 最終セルで大破警告を出さない
        /// </summary>
        [TestMethod]
        public void NotWarnDamagedShipInLastCell()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "taiha_001");
            PAssert.That(() => sniffer.GetShipStatuses(0)[2].NowHp == 2);
            PAssert.That(() => !sniffer.BadlyDamagedShips.Any());
        }

        /// <summary>
        /// 道中で大破警告を出す
        /// </summary>
        [TestMethod]
        public void WarnDamagedShip()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "taiha_002");
            PAssert.That(() => sniffer.GetShipStatuses(0)[2].NowHp == 1);
            PAssert.That(() => sniffer.BadlyDamagedShips.Any());
        }

        /// <summary>
        /// 連合艦隊の第二艦隊の旗艦の大破警告を出さない
        /// </summary>
        [TestMethod]
        public void NotWarnDamagedFlagshipInGuardFleet()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "taiha_003");
            PAssert.That(() => sniffer.GetShipStatuses(1)[0].NowHp == 12);
            PAssert.That(() => !sniffer.BadlyDamagedShips.Any());
        }

        /// <summary>
        /// 熟練度込みの制空値を正しく計算する
        /// </summary>
        [TestMethod]
        public void FighterPowerWithBonus()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "fighterpower_001");
            PAssert.That(() => sniffer.GetFighterPower(0).SequenceEqual(new[] {156, 159}));
            SniffLogFile(sniffer, "fighterpower_002");
            PAssert.That(() => sniffer.GetFighterPower(0).SequenceEqual(new[] {140, 143}), "全滅したスロットがある");
        }

        /// <summary>
        /// 改修効果込みの制空値を正しく計算する
        /// </summary>
        [TestMethod]
        public void FighterPowerWithImprovement()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "fighterpower_003");
            PAssert.That(() => sniffer.GetFighterPower(0).SequenceEqual(new[] {135, 135}));
        }

        /// <summary>
        /// 基地航空隊の制空値を正しく計算する
        /// </summary>
        [TestMethod]
        public void FighterPowerOfBaseAirCorps()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "baseaircorps_001");
            PAssert.That(() => sniffer.BaseAirCorps[0].AirCorps[0].FighterPower[1] == 301);
            sniffer.BaseAirCorps[0].AirCorps[0].Action = 2; // 防空
            PAssert.That(() => sniffer.BaseAirCorps[0].AirCorps[0].FighterPower[1] == 320);
        }

        /// <summary>
        /// 基地航空隊の防空時の偵察機補正を含む制空値を計算する
        /// </summary>
        [TestMethod]
        public void FighterPowerWithReconBonus()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "baseaircorps_002");
            PAssert.That(() => sniffer.BaseAirCorps[0].AirCorps[2].FighterPower[0] == 353);
        }

        /// <summary>
        /// 陸上攻撃機の熟練度を制空値に反映させる
        /// </summary>
        [TestMethod]
        public void FighterPowerOfLandBasedAttackAircraft()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "baseaircorps_003");
            PAssert.That(() => sniffer.BaseAirCorps[1].AirCorps[0].FighterPower[0] == 121);
        }

        /// <summary>
        /// 敵艦隊の制空値を計算する
        /// </summary>
        [TestMethod]
        public void EnemyFighterPower()
        {
            var sniffer1 = new Sniffer{UseOldEnemyId = true};
            SniffLogFile(sniffer1, "airbattle_002");
            PAssert.That(() => sniffer1.Battle.EnemyFighterPower.AirCombat == 28);

            var sniffer2 = new Sniffer {UseOldEnemyId = true};
            SniffLogFile(sniffer2, "both_combined_001");
            PAssert.That(() => sniffer2.Battle.EnemyFighterPower.AirCombat == 389);
        }

        /// <summary>
        /// マップ索敵の判定式(33)を正しく計算する
        /// </summary>
        [TestMethod]
        public void LineOfSight()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "lineofsight_001");
            PAssert.That(() => Math.Abs(sniffer.GetFleetLineOfSights(0, 1) - 39.45) < 0.01);
            PAssert.That(() => Math.Abs(sniffer.GetFleetLineOfSights(0, 3) - 115.19) < 0.01);
            PAssert.That(() => Math.Abs(sniffer.GetFleetLineOfSights(0, 4) - 153.06) < 0.01);
            SniffLogFile(sniffer, "lineofsight_002");
            PAssert.That(() => Math.Abs(sniffer.GetFleetLineOfSights(0, 1) - -25.10) < 0.01, "艦隊に空きがある");
        }

        /// <summary>
        /// 触接開始率を正しく計算する
        /// </summary>
        [TestMethod]
        public void ContactTriggerRage()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "contact_001");
            PAssert.That(() => Math.Abs(sniffer.GetContactTriggerRate(0) - 1.8182) < 0.0001);
        }

        /// <summary>
        /// 空襲戦を正しく処理する
        /// </summary>
        [TestMethod]
        public void LdAirBattle()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "ld_airbattle_001");
            AssertEqualBattleResult(sniffer, new[] {31, 59, 62, 14, 36, 47});
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.A);
        }

        /// <summary>
        /// 連合艦隊の空襲戦を正しく処理する
        /// </summary>
        [TestMethod]
        public void CombinedLdAirBattle()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "ld_airbattle_002");
            AssertEauqlCombinedResult(sniffer, new[] {31, 77, 77, 37, 41, 39}, new[] {45, 59, 64, 35, 36, 10});
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.A);
        }

        /// <summary>
        /// TPを正しく計算する
        /// </summary>
        [TestMethod]
        public void TransportPoint()
        {
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "transportpoint_001");
            ItemSpec.IncreaceLandPowerTp = false;
            PAssert.That(() => (int)sniffer1.GetShipStatuses(0).Sum(s => s.TransportPoint) == 27);
            ItemSpec.IncreaceLandPowerTp = true;
            PAssert.That(() => (int)sniffer1.GetShipStatuses(0).Sum(s => s.TransportPoint) == 37, "陸上戦力揚陸時");

            var sniffer2 = new Sniffer();
            SniffLogFile(sniffer2, "transportpoint_002");
            PAssert.That(() => (int)sniffer2.GetShipStatuses(0).Sum(s => s.TransportPoint) == 19, "鬼怒改二+特大発+おにぎり");
        }

        /// <summary>
        /// 敵の連合艦隊に対応する
        /// </summary>
        [TestMethod]
        public void EnemyCombinedBattle()
        {
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "ec_battle_001");
            PAssert.That(() => sniffer1.Battle.ResultRank == BattleResultRank.S, "昼戦のみ");

            var sniffer2 = new Sniffer();
            SniffLogFile(sniffer2, "ec_battle_002");
            PAssert.That(() => sniffer2.Battle.ResultRank == BattleResultRank.S, "夜戦込み");

            var sniffer3 = new Sniffer();
            SniffLogFile(sniffer3, "ec_battle_003");
            PAssert.That(() => sniffer3.Battle.ResultRank == BattleResultRank.A, "護衛を撃ちもらす");
        }


        /// <summary>
        /// 双方連合艦隊の戦闘に対応する
        /// </summary>
        [TestMethod]
        public void BothCombinedBattle()
        {
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "both_combined_water_001");
            PAssert.That(() => sniffer1.Battle.ResultRank == BattleResultRank.A);

            var sniffer2 = new Sniffer();
            SniffLogFile(sniffer2, "both_combined_001");
            PAssert.That(() => sniffer2.Battle.ResultRank == BattleResultRank.S);

            // 護衛艦隊に雷撃を受ける場合
            var sniffer3 = new Sniffer();
            SniffLogFile(sniffer3, "both_combined_002");
            PAssert.That(() => sniffer2.Battle.ResultRank == BattleResultRank.S);
            PAssert.That(
                () => sniffer3.GetShipStatuses(0).Select(s => s.NowHp).SequenceEqual(new[] {89, 86, 87, 78, 87, 57}));
            PAssert.That(
                () => sniffer3.GetShipStatuses(1).Select(s => s.NowHp).SequenceEqual(new[] {46, 37, 14, 59, 49, 63}));
        }

        /// <summary>
        /// 噴式強襲に対応する
        /// </summary>
        [TestMethod]
        public void Injection()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "injection_001");
            PAssert.That(() => sniffer.Battle.ResultRank == BattleResultRank.P);
        }

        /// <summary>
        /// 対空砲火のパラメータを計算する
        /// </summary>
        [TestMethod]
        public void AntiAirFire()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "antiairfire_001");
            var ships = sniffer.GetShipStatuses(0);
            PAssert.That(() => ships.Sum(ship => ship.EffectiveAntiAirForFleet) == 88);
            PAssert.That(
                () =>
                    ships.Select(ship => ship.EffectiveAntiAirForShip)
                        .SequenceEqual(new[] {92, 90, 88, 228, 146, 226}));
        }

        /// <summary>
        /// 航空戦の結果を処理する
        /// </summary>
        [TestMethod]
        public void AirBattleResult()
        {
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "battle_001");
            PAssert.That(() => sniffer1.Battle.AirBattleResults.Count == 0);

            var sniffer2 = new Sniffer();
            SniffLogFile(sniffer2, "ec_battle_001");
            PAssert.That(() => sniffer2.Battle.AirBattleResults.Count == 4);

            var sniffer3 = new Sniffer();
            SniffLogFile(sniffer3, "battle_007");
            PAssert.That(() => sniffer3.Battle.AirBattleResults.Count == 1, "stage2がない場合");
        }

        /// <summary>
        /// 2-5をクリアしたときの特別戦果を反映する
        /// </summary>
        [TestMethod]
        public void ExMapBattleResult()
        {
            var sniffer0 = new Sniffer();
            sniffer0.ExMap.ClearClearStatus();
            SniffLogFile(sniffer0, "eo_001");
            PAssert.That(() => sniffer0.ExMap.Achievement == 100, "ほかのマップの情報なし");
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "eo_001");
            // すでに3-5をクリアしているので合計で250
            PAssert.That(() => sniffer1.ExMap.Achievement == 250, "ほかのマップの情報あり");
        }

        /// <summary>
        /// 1-6をクリアしたときの特別戦果を反映する
        /// </summary>
        [TestMethod]
        public void ExMapMapNext()
        {
            var sniffer0 = new Sniffer();
            sniffer0.ExMap.ClearClearStatus();
            SniffLogFile(sniffer0, "eo_002");
            PAssert.That(() => sniffer0.ExMap.Achievement == 75, "ほかのマップの情報なし");
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "eo_002");
            // 5-5以外クリアしているので合計で400
            PAssert.That(() => sniffer1.ExMap.Achievement == 400, "ほかのマップの情報あり");
        }

        /// <summary>
        /// 編成で空き番号を使ったローテートを正しく反映する
        /// </summary>
        [TestMethod]
        public void RotateFleetMember()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "deck_001");
            var result = sniffer.GetDeck(0);
            PAssert.That(() => new[] {756, 17204, 6156, 28806, 1069, -1}.SequenceEqual(result));
        }

        /// <summary>
        /// ドラッグ＆ドロップで離れた空き番号を使って編成をローテートする
        /// </summary>
        [TestMethod]
        public void RotateFleetMemberWithDragAndDrop()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "deck_005");
            var result = sniffer.GetDeck(0);
            PAssert.That(() => new[] {57391, 50, 24475, 113, -1, -1}.SequenceEqual(result));
        }

        /// <summary>
        /// 編成で艦隊に配置ずみの艦娘を交換する
        /// </summary>
        [TestMethod]
        public void ExchangeFleetMember()
        {
            var sniffer = new Sniffer();

            SniffLogFile(sniffer, "deck_002");
            var result0 = sniffer.GetDeck(0);
            PAssert.That(() => new[] {1069, 6156, 756, 3223, -1, -1}.SequenceEqual(result0), "編成で艦隊内で艦娘と交換する");

            SniffLogFile(sniffer, "deck_003");
            var result10 = sniffer.GetDeck(0);
            var result11 = sniffer.GetDeck(1);
            PAssert.That(() => new[] {1069, 6156, 14258, 3223, -1, -1}.SequenceEqual(result10) &&
                               new[] {101, 4487, 756, 14613, 28806, -1}.SequenceEqual(result11), "002に続いて艦隊をまたがって交換する");

            SniffLogFile(sniffer, "deck_004");
            var result20 = sniffer.GetDeck(0);
            var result21 = sniffer.GetDeck(1);
            PAssert.That(() => new[] {1069, 6156, 14258, 3223, 756, -1}.SequenceEqual(result20) &&
                               new[] {101, 4487, 14613, 28806, -1, -1}.SequenceEqual(result21),
                "003に続いて空き番号にほかの艦隊の艦娘を配置する");
        }

        /// <summary>
        /// 編成展開を正しく反映する
        /// </summary>
        [TestMethod]
        public void PresetSelect()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "preset_001");
            var result = sniffer.GetDeck(0);
            PAssert.That(() => new[] {50510, 632, 39843, 113, 478, 47422}.SequenceEqual(result));
        }

        /// <summary>
        /// 拡張した編成記録枠にすぐに記録してもエラーにならない
        /// </summary>
        [TestMethod]
        public void PresetExpand()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "preset_002");
        }

        /// <summary>
        /// 装備の交換を正しく反映する
        /// </summary>
        [TestMethod]
        public void SlotExchange()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "slot_exchange_001");
            var result = sniffer.GetShipStatuses(0)[0].Slot.Select(item => item.Id);
            PAssert.That(() => new[] {26096, 30571, 77694, 61383, -1}.SequenceEqual(result));
        }

        /// <summary>
        /// 近代化改修の結果をすぐに反映する
        /// </summary>
        [TestMethod]
        public void PowerupResult()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "powerup_001");
            PAssert.That(() => Math.Abs(sniffer.GetShipStatuses(0)[0].EffectiveFirepower - 30) < 0.0001);
        }

        /// <summary>
        /// 近代化改修が二重に行われた場合に対応する
        /// </summary>
        [TestMethod]
        public void DuplicatedPowerup()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "powerup_002");
            PAssert.That(() => sniffer.Item.NowShips == 218);
        }

        /// <summary>
        /// 改修による資材の減少をすぐに反映する
        /// </summary>
        [TestMethod]
        public void ConsumptionByRemodelSlot()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "remodel_slot_001");
            PAssert.That(() => sniffer.Material.Current
                .SequenceEqual(new[] {25292, 25570, 25244, 41113, 1405, 1525, 2137, 8}));
        }

        /// <summary>
        /// 装備の数を正しく数える
        /// </summary>
        [TestMethod]
        public void CountItem()
        {
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "createitem_001");
            PAssert.That(() => sniffer1.Item.NowEquips == 900);
            var sniffer2 = new Sniffer();
            SniffLogFile(sniffer2, "createship_001");
            PAssert.That(() => sniffer2.Item.NowEquips == 904);
        }

        /// <summary>
        /// 解体された艦の装備数を全体の装備数から引く
        /// </summary>
        [TestMethod]
        public void CountItemsOnDeletedShip()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "destroyship_001");
            PAssert.That(() => sniffer.Item.NowEquips == 879, "ドロップしたばかりの艦娘を解体する場合");
        }

        /// <summary>
        /// 装備数の超過を警告する
        /// </summary>
        [TestMethod]
        public void WarnItemCount()
        {
            var sniffer1 = new Sniffer();
            SniffLogFile(sniffer1, "item_count_001");
            PAssert.That(() => sniffer1.Item.RingEquips, "出撃から母港に戻ったとき");
            var sniffer2 = new Sniffer();
            SniffLogFile(sniffer2, "item_count_002");
            PAssert.That(() => sniffer2.Item.RingEquips, "ログインしたとき");
        }

        /// <summary>
        /// ドロップ艦のいる出撃から戻ったときに、slot_itemを受信する前にItemListを取得しても例外を出さない
        /// </summary>
        [TestMethod]
        public void GetItemListBeforeReceivingSlotItem()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "dropship_001");
            PAssert.That(() => sniffer.ItemList[0].Holder.Id == 250);
        }

        /// <summary>
        /// 装備の所持者を設定する
        /// </summary>
        [TestMethod]
        public void SetItemHolder()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "itemholder_001");
            var itemIds = new HashSet<int> {75298, 76572, 82725, 90213, 5910};
            var items = sniffer.ItemList.Where(status => itemIds.Contains(status.Id));
            PAssert.That(() => items.All(x => x.Holder.Id == 861));
        }

        /// <summary>
        /// 資材の変動を正しく反映する
        /// </summary>
        [TestMethod]
        public void MaterialChanges()
        {
            var sniffer1 = new Sniffer();
            var result1 = new List<int[]>();
            SniffLogFile(sniffer1, "material_001", sn =>
            {
                var cur = sn.Material.Current;
                if (result1.Count == 0)
                {
                    result1.Add(cur);
                }
                else
                {
                    if (!result1.Last().SequenceEqual(cur))
                        result1.Add(cur);
                }
            });
            var expected1 = new List<int[]>
            {
                new[] {0, 0, 0, 0, 0, 0, 0, 0},
                new[] {26178, 26742, 21196, 33750, 1426, 1574, 2185, 10},
                new[] {26178, 26842, 21226, 33750, 1426, 1574, 2185, 10},
                new[] {28951, 29493, 24945, 35580, 1426, 1574, 2185, 10},
                new[] {26074, 26616, 21068, 33700, 1426, 1572, 2183, 10},
                new[] {26171, 26721, 21175, 33750, 1426, 1574, 2185, 10},
                new[] {27023, 27829, 28136, 42404, 1404, 1521, 2142, 15},
                new[] {31208, 29819, 29714, 42345, 1407, 1530, 2155, 13},
                new[] {24595, 25353, 18900, 32025, 1427, 1576, 2187, 10},
                new[] {24515, 25353, 18749, 32025, 1427, 1575, 2187, 10},
                new[] {23463, 24964, 17284, 31765, 1427, 1572, 2187, 10},
                new[] {23463, 25064, 17314, 31765, 1427, 1572, 2187, 10}
            };
            PAssert.That(() => SequenceOfSequenceEqual(expected1, result1));

            var sniffer2 = new Sniffer();
            var result2 = new List<int[]>();
            SniffLogFile(sniffer2, "material_002", sn =>
            {
                var cur = sn.Material.Current;
                if (result2.Count == 0)
                {
                    result2.Add(cur);
                }
                else
                {
                    if (!result2.Last().SequenceEqual(cur))
                        result2.Add(cur);
                }
            });
            var expected2 = new List<int[]>
            {
                new[] {0, 0, 0, 0, 0, 0, 0, 0},
                new[] {201649, 189713, 261490, 123227, 2743, 2828, 3000, 44},
                new[] {201649, 189714, 261491, 123227, 2743, 2828, 3000, 44},
                new[] {201650, 189718, 261500, 123227, 2743, 2828, 3000, 44}
            };
            PAssert.That(() => SequenceOfSequenceEqual(expected2, result2));
        }

        /// <summary>
        /// 基地航空隊における資材の変動を反映する
        /// </summary>
        [TestMethod]
        public void MaterialChangesInAirCorps()
        {
            var sniffer3 = new Sniffer();
            var result3 = new List<int[]>();
            SniffLogFile(sniffer3, "material_003", sn =>
            {
                var cur = sn.Material.Current;
                if (result3.Count == 0)
                {
                    result3.Add(cur);
                }
                else
                {
                    if (!result3.Last().SequenceEqual(cur))
                        result3.Add(cur);
                }
            });
            var expected3 = new List<int[]>
            {
                new[] {0, 0, 0, 0, 0, 0, 0, 0},
                new[] {288194, 282623, 299496, 295958, 3000, 2968, 2997, 7},
                new[] {288185, 282623, 299496, 295943, 3000, 2968, 2997, 7},
                new[] {288161, 282623, 299496, 295903, 3000, 2968, 2997, 7},
            };
            PAssert.That(() => SequenceOfSequenceEqual(expected3, result3), "航空機の補充");

            var sniffer4 = new Sniffer();
            var result4 = new List<int[]>();
            SniffLogFile(sniffer4, "material_004", sn =>
            {
                var cur = sn.Material.Current;
                if (result4.Count == 0)
                {
                    result4.Add(cur);
                }
                else
                {
                    if (!result4.Last().SequenceEqual(cur))
                        result4.Add(cur);
                }
            });
            var expected4 = new List<int[]>
            {
                new[] {0, 0, 0, 0, 0, 0, 0, 0},
                new[] {261012, 252252, 298492, 279622, 3000, 2842, 3000, 22},
                new[] {261012, 252252, 298492, 279538, 3000, 2842, 3000, 22},
                new[] {261012, 252252, 298492, 279454, 3000, 2842, 3000, 22},
            };
            PAssert.That(() => SequenceOfSequenceEqual(expected4, result4), "航空機の配備");
        }

        private bool SequenceOfSequenceEqual<T>(IEnumerable<IEnumerable<T>> a, IEnumerable<IEnumerable<T>> b)
        {
            var aa = a.ToArray();
            var bb = b.ToArray();
            if (aa.Length != bb.Length)
                return false;
            return aa.Zip(bb, (x, y) => x.SequenceEqual(y)).All(x => x);
        }

        /// <summary>
        /// 海域ゲージ情報を作成する
        /// </summary>
        [TestMethod]
        public void MapInfo()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "mapinfo_001");
            PAssert.That(() => sniffer.MiscText ==
                               "[海域ゲージ]\r\n" +
                               "1-6 : 撃破 3/7\r\n5-5 : 撃破 0/5\r\n6-3 : 撃破 0/4\r\n" +
                               "31-1 : HP 0/750\r\n31-2 : HP 0/1050\r\n31-3 : HP 0/2100\r\n31-4 : HP 0/3500\r\n31-5 : HP 1/2450\r\n",
                "最初の海域ゲージ情報");
            SniffLogFile(sniffer, "mapinfo_002");
            PAssert.That(() => sniffer.MiscText ==
                               "[海域ゲージ情報]\r\n 海域選択画面に進むと表示します。\r\n[演習情報]\r\n 演習相手を選ぶと表示します。",
                "戦闘から戻ったら案内を表示");
            SniffLogFile(sniffer, "mapinfo_003");
            PAssert.That(() => sniffer.MiscText ==
                               "[海域ゲージ]\r\n" +
                               "1-6 : 撃破 4/7\r\n5-5 : 撃破 0/5\r\n6-3 : 撃破 0/4\r\n" +
                               "31-1 : HP 0/750\r\n31-2 : HP 0/1050\r\n31-3 : HP 0/2100\r\n31-4 : HP 0/3500\r\n31-5 : HP 1/2450\r\n",
                "更新された海域ゲージ情報");
        }

        /// <summary>
        /// 演習相手の情報を作成する
        /// </summary>
        [TestMethod]
        public void PractiveEnemyInfo()
        {
            var sniffer = new Sniffer();
            SniffLogFile(sniffer, "practice_enemyinfo_001");
            PAssert.That(() => sniffer.MiscText ==
                               "[演習情報]\r\n敵艦隊名 : 大銀河地球連邦　第7艦隊\r\n獲得経験値 : 570\r\nS勝利 : 684",
                "演習相手の情報");
            SniffLogFile(sniffer, "practice_enemyinfo_002");
            PAssert.That(() => sniffer.MiscText ==
                               "[海域ゲージ情報]\r\n 海域選択画面に進むと表示します。\r\n[演習情報]\r\n 演習相手を選ぶと表示します。",
                "演習から戻ったら案内を表示");
        }
    }
}